﻿using System;
using System.Reflection;

namespace UMK.WzorceStrukturalne
{
    #region Klient i klasy używane przez niego
    class Punkt
    {
        public float X, Y;

        public Punkt(float x, float y)
        {
            X = x;
            Y = y;
        }
    }

    //target, interfejs używany przez clienta
    interface IWielokątForemny
    {
        Punkt Środek { get; }
        int LiczbaBoków { get; }
        float DługośćBoków { get; }

        void WyświetlParametry();
        void WyświetlNazwęFigury();
        float ObliczObwód();
        float ObliczPole();
    }

    abstract class WielokątForemny : IWielokątForemny
    {
        public Punkt Środek { get; private set; }
        public int LiczbaBoków { get; private set; }
        public float DługośćBoków { get; private set; }

        protected WielokątForemny(Punkt środek, int liczbaBoków, float długośćBoków)
        {
            this.Środek = środek;
            this.LiczbaBoków = liczbaBoków;
            this.DługośćBoków = długośćBoków;
        }

        public virtual void WyświetlParametry()
        {
            Console.WriteLine("srodek: (" + Środek.X + "," + Środek.Y + "), liczba bokow: " + LiczbaBoków + ", dlugosc bokow: " + DługośćBoków);
        }

        public abstract void WyświetlNazwęFigury();

        public float ObliczObwód() //bez modyfikatora virtual - implementacja "ostateczna"
        {
            return LiczbaBoków * DługośćBoków;
        }

        public abstract float ObliczPole();
    }

    //przykład klasy zgodnej z targetem
    class TrójkątForemny : WielokątForemny
    {

        public TrójkątForemny(Punkt środek, float długośćBoków)
            : base(środek, 3, długośćBoków)
        { }

        public override void WyświetlNazwęFigury()
        {
            Console.Write("Trojkat rownoboczny");
        }

        public override float ObliczPole()
        {
            return DługośćBoków * DługośćBoków * (float)(Math.Sqrt(3.0) / 4.0);
        }
    };

    //client
    class Klient
    {
        public static void WyświetlInformacjeOFigurze(IWielokątForemny wielokątForemny)
        {
            wielokątForemny.WyświetlNazwęFigury();
            Console.WriteLine(":");
            wielokątForemny.WyświetlParametry();
            Console.WriteLine("Obwod: " + wielokątForemny.ObliczObwód());
            Console.WriteLine("Pole: " + wielokątForemny.ObliczPole());
            Console.WriteLine();
        }
    }
    #endregion

    #region Kod niepasujący, którego jednak chcemy użyć
    //adaptee, klasa adaptowana
    class Prostokąt
    {
        protected Punkt p1; //lewy-górny róg
        protected Punkt p2; //prawy-dolny róg	

        protected float szerokość() { return p2.X - p1.X; }
        protected float wysokość() { return p2.Y - p1.Y; }

        public Prostokąt(Punkt p1, Punkt p2)
        {
            this.p1 = p1;
            this.p2 = p2;
        }

        public void WyświetlNazwęFigury() //inna nazwa funkcji
        {
            Console.Write("Prostokat");
        }

        public void WyświetlParametry() //wyświetla inny rodzaj parametrów
        {
            Console.WriteLine("lewa krawędź: " + p1.X + ", górna krawędź: " + p1.Y + ", szerokość: " + szerokość() + ", wysokość: " + wysokość());
        }

        public float ObliczPole()
        {
            return szerokość() * wysokość();
        }

        public float _ObliczObwód()
        {
            return 2 * szerokość() + 2 * wysokość();
        }
    }
    #endregion

    #region Nakładka dostosowująca kod niepasującego do targetu
    class ProstokątForemny : WielokątForemny
    {
        private static Punkt obliczP1(Punkt środek, float długośćBoków)
        {
            return new Punkt(środek.X - długośćBoków / 2, środek.Y - długośćBoków / 2);
        }

        private static Punkt obliczP2(Punkt środek, float długośćBoków)
        {
            return new Punkt(środek.X + długośćBoków / 2, środek.Y + długośćBoków / 2);
        }

        /*
        public ProstokątForemny(Punkt środek, float długośćBoków)
            : base(obliczP1(środek, długośćBoków), obliczP2(środek, długośćBoków))
        {
            Środek = środek;
            DługośćBoków = długośćBoków;
        }

        public Punkt Środek { get; private set; }

        public int LiczbaBoków => 4;

        public float DługośćBoków { get; private set; }

        public float ObliczObwód()
        {
            return _ObliczObwód();
        }
        */

        class _Prostokąt : Prostokąt
        {
            public _Prostokąt(Punkt p1, Punkt p2)
                :base(p1,p2)
            {
            }

            public Punkt P1
            {
                get
                {
                    return p1;
                }
            }

            public Punkt P2
            {
                get
                {
                    return p2;
                }
            }
        }

        public static object getValueFromMemberInfo(MemberInfo memberInfo, object forObject)
        {
            switch (memberInfo.MemberType)
            {
                case MemberTypes.Field:
                    return ((FieldInfo)memberInfo).GetValue(forObject);
                case MemberTypes.Property:
                    return ((PropertyInfo)memberInfo).GetValue(forObject);
                default:
                    throw new NotImplementedException();
            }
        }

        private static _Prostokąt zróbOdbicie(Prostokąt prostokąt)
        {
            Punkt p1 = (Punkt)getValueFromMemberInfo((MemberInfo)(prostokąt.GetType().GetMember("p1", BindingFlags.Instance | BindingFlags.NonPublic).GetValue(0)), prostokąt);
            Punkt p2 = (Punkt)getValueFromMemberInfo((MemberInfo)(prostokąt.GetType().GetMember("p2", BindingFlags.Instance | BindingFlags.NonPublic).GetValue(0)), prostokąt);
            return new _Prostokąt(p1, p2);
        }

        private static _Prostokąt klonformuj(Prostokąt prostokąt)
        {
            //_Prostokąt _prostokąt = (_Prostokąt)prostokąt;
            _Prostokąt _prostokąt = zróbOdbicie(prostokąt);
            return _prostokąt;
        }

        private static bool czyProstokątNieJestForemny(Prostokąt prostokąt)
        {
            /*
            double a = prostokąt._ObliczObwód() / 4;
            double P = a * a;
            return P != prostokąt.ObliczPole();
            */

            _Prostokąt _prostokąt = klonformuj(prostokąt);
            float a = Math.Abs(_prostokąt.P1.X - _prostokąt.P2.X);
            float b = Math.Abs(_prostokąt.P1.Y - _prostokąt.P2.Y);
            return a == b;
        }

        private static Punkt obliczŚrodek(Prostokąt prostokąt)
        {
            _Prostokąt _prostokąt = klonformuj(prostokąt);
            float x = (_prostokąt.P1.X + _prostokąt.P2.X) / 2;
            float y = (_prostokąt.P1.Y + _prostokąt.P2.Y) / 2;
            return new Punkt(x, y);
        }

        private static float obliczDługośćBoków(Prostokąt prostokąt)
        {
            _Prostokąt _prostokąt = klonformuj(prostokąt);
            float a = Math.Abs(_prostokąt.P1.X - _prostokąt.P2.X);            
            return a;
        }

        private Prostokąt prostokąt;

        public ProstokątForemny(Punkt środek, float długośćBoków)
            :base(środek,4,długośćBoków)
        {
            prostokąt = new Prostokąt(obliczP1(środek, długośćBoków), obliczP2(środek, długośćBoków));
        }

        public ProstokątForemny(Prostokąt prostokąt)
            : base(obliczŚrodek(prostokąt), 4, obliczDługośćBoków(prostokąt))
        {
            if (!czyProstokątNieJestForemny(prostokąt)) throw new Exception("Prostokąt nie jest kwadratem");
            this.prostokąt = prostokąt; //TODO: rozważyć klonowanie
        }

        public override void WyświetlNazwęFigury()
        {
            System.Console.WriteLine("Kwadrat");
        }

        public override float ObliczPole()
        {
            return prostokąt.ObliczPole();
        }
    }
    #endregion
}
